import { singleton } from 'decorators/singleton'
import assert from 'assert'
import { HttpRetryProvider } from 'chain/HttpRetryProvider'
import { AllChains } from 'chain/allchain'
import Web3 from 'web3'
import { WalletReactor } from 'chain/WalletReactor'
import { ConfirmQueue } from 'queue/confirm.queue'

let chainCfgs = require('../../config/chains.json')

export interface IChainCfg {
  id: number
  key: string
  wallet: string
  mail: string
}

@singleton
export class ChainCache {
  public providers: Map<number, Web3> = new Map()
  public wallets: Map<number, WalletReactor> = new Map()
  public confirmQueues: Map<number, ConfirmQueue> = new Map()
  constructor() {}

  public async init(filename: string) {
    chainCfgs = require(`../../config/${filename}.json`)
    for (let cfg of chainCfgs) {
      let chainData = AllChains.find(o => o.id === cfg.id)
      assert(chainData, `chain id ${cfg.id} not found`)
      let web3 = new Web3(new HttpRetryProvider(chainData.rpc.split('|')))
      web3.eth.accounts.wallet.add(cfg.key)
      this.providers.set(cfg.id, web3)
      let wallet = new WalletReactor({
        web3: web3,
        address: cfg.wallet,
      })
      this.wallets.set(cfg.id, wallet)
      this.confirmQueues.set(cfg.id, new ConfirmQueue(web3))
    }
  }

  public get chainArray(): IChainCfg[] {
    return chainCfgs
  }

  public getWeb3(chainId: number) {
    if (!this.providers.has(chainId)) {
      throw new Error(`web3 for chain id ${chainId} not found`)
    }
    return this.providers.get(chainId)
  }

  public getWallet(chainId: number) {
    if (!this.wallets.has(chainId)) {
      throw new Error(`wallet for chain id ${chainId} not found`)
    }
    return this.wallets.get(chainId)
  }

  public getConfirmQueue(chainId: number) {
    if (!this.confirmQueues.has(chainId)) {
      throw new Error(`confirm queue for chain id ${chainId} not found`)
    }
    return this.confirmQueues.get(chainId)
  }

  public getInstances(chain: number, contract: string, abi: any) {
    let web3 = new ChainCache().getWeb3(chain)
    assert(web3, `web3 for chain id ${chain} not found`)
    let account = web3.eth.accounts.wallet[0]
    assert(account, `account for chain id ${chain} not found`)
    return { instance: new web3.eth.Contract(abi, contract, { from: account.address }), account, web3 }
  }
}
