import { singleton } from 'decorators/singleton'
import logger from 'logger/logger'
import { TaskType } from 'models/RequestTask'
import { ConfirmQueue } from 'queue/confirm.queue'
import { TransactionReceipt } from 'web3-core'
import { ERC20Reactor } from './ERC20Reactor'
import { ERC721Reactor } from './ERC721Reactor'
import { DistributorReactor } from './DistributorReactor'
import { ChainCache } from 'cache/ChainCache'

@singleton
export class BlockChain {
  public erc20Reactor: ERC20Reactor
  public erc721Reactor: ERC721Reactor
  public distributorReactor: DistributorReactor
  public confirmQueue: ConfirmQueue
  public currentBlockNum: number = 0

  constructor() {
    new ChainCache()
    this.erc20Reactor = new ERC20Reactor()
    this.erc721Reactor = new ERC721Reactor()
    this.distributorReactor = new DistributorReactor()
  }

  public async getTxConfirms(chain: number, txhash: string) {
    let web3 = new ChainCache().getWeb3(chain)
    const receipt: TransactionReceipt = await web3.eth.getTransactionReceipt(txhash)
    const latest = await web3.eth.getBlockNumber()
    return latest - receipt.blockNumber + 1
  }

  public async updateCurrenBlockNum(chain: number) {
    let web3 = new ChainCache().getWeb3(chain)
    this.currentBlockNum = await web3.eth.getBlockNumber()
    // logger.debug(`update block num: ${this.currentBlockNum}`)
  }

  public async generateFunAbi(reqData: any) {
    let taskType = parseInt(reqData.type)
    reqData.encodeABI = true
    let abi
    switch (taskType) {
      case TaskType.MINT_FT:
        abi = await this.erc20Reactor.mint(reqData)
        break
      case TaskType.MINT_NFT:
        reqData.tokenId = reqData.tokenId || reqData.tokenid
        if (reqData.tokenId && !reqData.amount) {
          reqData.tokenIds = [reqData.tokenId]
          abi = await this.erc721Reactor.mintNFT(reqData)
        } else {
          abi = await this.erc721Reactor.mintBadge(reqData)
        }
        break
      case TaskType.TRANSFER_FT:
        abi = await this.erc20Reactor.transfer(reqData)
        break
      case TaskType.TRANSFER_NFT:
        reqData.tokenId = reqData.tokenId || reqData.tokenid
        abi = await this.erc721Reactor.transfer(reqData)
        break
      case TaskType.PUBLISH_AIRDROP_LIST:
        reqData.nftList = [reqData.tokenId || reqData.tokenid]
        abi = await this.distributorReactor.publishAirdropList(reqData)
        break
    }
    return abi
  }
}
